package MIC.module.algorithm;

import adf.agent.communication.MessageManager;
import adf.agent.develop.DevelopData;
import adf.agent.info.AgentInfo;
import adf.agent.info.ScenarioInfo;
import adf.agent.info.WorldInfo;
import adf.agent.module.ModuleManager;
import adf.agent.precompute.PrecomputeData;
import adf.component.module.algorithm.Clustering;
import adf.component.module.algorithm.PathPlanning;
import adf.component.module.algorithm.StaticClustering;
import adf.launcher.ConsoleOutput;
import com.google.common.collect.Iterables;

import MIC.util.Convertors;

import org.apache.commons.math3.random.MersenneTwister;
import org.apache.commons.math3.random.RandomGenerator;
import rescuecore2.misc.Pair;
import rescuecore2.misc.collections.LazyMap;
import rescuecore2.misc.geometry.Point2D;
import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.Entity;
import rescuecore2.worldmodel.EntityID;

import java.io.Serializable;
import java.util.*;

public class MICKMeans extends StaticClustering {
    private static final String KEY_CLUSTER_SIZE = "sample.clustering.size";
    private static final String KEY_CLUSTER_CENTER = "sample.clustering.centers";
    private static final String KEY_CLUSTER_ENTITY = "sample.clustering.entities.";
    private static final String KEY_ASSIGN_AGENT = "sample.clustering.assign";

    private int repeatPrecompute;
    private int repeatPreparate;
    
    private Convertors convertor;

    private Collection<StandardEntity> entities;

    private List<StandardEntity> centerList;
    private List<EntityID> centerIDs;
    private Map<Integer, List<StandardEntity>> clusterEntitiesList;
    private List<List<EntityID>> clusterEntityIDsList;

    private int clusterSize;

    private boolean assignAgentsFlag;

    private Map<EntityID, Set<EntityID>> shortestPathGraph;
    protected RandomGenerator random;

    protected List<StandardEntity> agents;
    protected Map<EntityID, Integer> agentIDClusterIndexMapping;
    protected List<StandardEntityURN> clusterTypes;

    protected PathPlanning pathPlanning;

    public MICKMeans(AgentInfo ai, WorldInfo wi, ScenarioInfo si, ModuleManager moduleManager, DevelopData developData) {
        super(ai, wi, si, moduleManager, developData);
        this.repeatPrecompute = developData.getInteger("MICKMeans.repeatPrecompute", 1);
        this.repeatPreparate = developData.getInteger("MICKMeans.repeatPreparate", 1);
        this.clusterSize = developData.getInteger("MICKMeans.clusterSize", 1);
        this.assignAgentsFlag = developData.getBoolean("MICKMeans.assignAgentsFlag", true);
        int randomNumberSeed = developData.getInteger("MICKMeans.randomSeed", 1);

        this.pathPlanning = moduleManager.getModule("MICKMeans.PathPlanning", "MIC.module.algorithm.MICPathPlanning");

        this.clusterEntityIDsList = new ArrayList<>();
        this.centerIDs = new ArrayList<>();
        this.clusterEntitiesList = new HashMap<>();
        this.centerList = new ArrayList<>();

        this.entities = wi.getEntitiesOfType(
                StandardEntityURN.ROAD,
                StandardEntityURN.HYDRANT,
                StandardEntityURN.BUILDING,
                StandardEntityURN.REFUGE,
                StandardEntityURN.GAS_STATION,
                StandardEntityURN.AMBULANCE_CENTRE,
                StandardEntityURN.FIRE_STATION,
                StandardEntityURN.POLICE_OFFICE
        );

        clusterTypes = new ArrayList<>();
        clusterTypes.add(StandardEntityURN.ROAD);
        clusterTypes.add(StandardEntityURN.HYDRANT);
        clusterTypes.add(StandardEntityURN.BUILDING);
        clusterTypes.add(StandardEntityURN.REFUGE);
        clusterTypes.add(StandardEntityURN.GAS_STATION);
        clusterTypes.add(StandardEntityURN.AMBULANCE_CENTRE);
        clusterTypes.add(StandardEntityURN.FIRE_STATION);
        clusterTypes.add(StandardEntityURN.POLICE_OFFICE);

        random = new MersenneTwister(randomNumberSeed);

        StandardEntityURN agentType = getAgentType();
        agents = new ArrayList<>(wi.getEntitiesOfType(agentType));
        clusterSize = agents.size();
        agentIDClusterIndexMapping = new HashMap<>();
        
        convertor = new Convertors(this.worldInfo);

    }

    @Override
    public Clustering updateInfo(MessageManager messageManager) {
        super.updateInfo(messageManager);
        if(this.getCountUpdateInfo() >= 2) {
            return this;
        }
        this.centerList.clear();
        this.clusterEntitiesList.clear();
        return this;
    }

    @Override
    public Clustering precompute(PrecomputeData precomputeData) {
        super.precompute(precomputeData);
        if(this.getCountPrecompute() >= 2) {
            return this;
        }
        return this;
    }

    @Override
    public Clustering resume(PrecomputeData precomputeData) {
        super.resume(precomputeData);
        if(this.getCountResume() >= 2) {
            return this;
        }
        preparate();
        return this;
    }

    @Override
    public Clustering preparate() {
        super.preparate();
        if(this.getCountPreparate() >= 2) {
            return this;
        }
        repeatPreparate = 20;
        if (getAgentType() == StandardEntityURN.POLICE_FORCE) {
            clusterTypes = new ArrayList<>(2);
            clusterTypes.add(StandardEntityURN.ROAD);
            clusterTypes.add(StandardEntityURN.HYDRANT);

            entities = worldInfo.getEntitiesOfType(StandardEntityURN.ROAD,
                    StandardEntityURN.HYDRANT);
            calcStandard(repeatPreparate);
            connectAllClusterEntities();
        } else {
            calcStandard(repeatPreparate);
        }
        debugDrawing();
        this.entities = null;
        return this;
    }

    protected void debugDrawing() {
        Map<Integer, List<Integer>> clusters = new HashMap<>();
        int index = 0;
        for (List<EntityID> cluster : clusterEntityIDsList) {
            List<Integer> clusterEntities = new ArrayList<>(cluster.size());
            for (EntityID id : cluster) {
                clusterEntities.add(id.getValue());
            }
            clusters.put(index, clusterEntities);
            index++;
        }
    }

    @Override
    public int getClusterNumber() {
        return this.clusterSize;
    }

    @Override
    public int getClusterIndex(StandardEntity entity) {
        return this.getClusterIndex(entity.getID());
    }

    @Override
    public int getClusterIndex(EntityID id) {

        if (agentIDClusterIndexMapping.containsKey(id)) {
            return agentIDClusterIndexMapping.get(id);
        } else {
            return -1;
        }
    }

    @Override
    public Collection<StandardEntity> getClusterEntities(int index) {
        if (clusterEntitiesList.containsKey(index)) {
            List<StandardEntity> result = this.clusterEntitiesList.get(index);
            if(result == null || result.isEmpty()) {
                List<EntityID> list = this.clusterEntityIDsList.get(index);
                result = new ArrayList<>(list.size());
                for(int i = 0; i < list.size(); i++) {
                    result.add(i, this.worldInfo.getEntity(list.get(i)));
                }
                this.clusterEntitiesList.put(index, result);
            }
            return result;
        } else {
            int objectDiameter = 100000;
            return getEntitiesInRange(objectDiameter, clusterTypes);
        }
    }

    protected List<StandardEntity> getEntitiesInRange(int range, List<StandardEntityURN> types) {
        List<StandardEntity> result = new ArrayList<>();
        Collection<EntityID> newEntities = worldInfo.getObjectIDsInRange((int)agentInfo.getX(), (int)agentInfo.getY(), range);
        for (EntityID id : newEntities) {
            StandardEntity se = worldInfo.getEntity(id);
            if (se != null) {
                StandardEntityURN myType = se.getStandardURN();
                for (StandardEntityURN type : types) {
                    if (myType.toString().equals(type.toString())) {
                        result.add(se);
                        break;
                    }
                }
            }
        }
        return result;
    }

    @Override
    public Collection<EntityID> getClusterEntityIDs(int index) {
        if (index < clusterEntityIDsList.size() && index >= 0)
            return this.clusterEntityIDsList.get(index);
        else {
            int range = 100000;
            List<StandardEntity> entities = getEntitiesInRange(range, clusterTypes);
            List<EntityID> entityIDs = new ArrayList<>(entities.size());
            for (StandardEntity se : entities) {
                entityIDs.add(se.getID());
            }
            return entityIDs;
        }
    }

    @Override
    public Clustering calc() {
        return this;
    }

    protected int agentDebugID = 2011514986;
    private void calcStandard(int repeat) {
        this.initShortestPath(this.worldInfo);

        List<StandardEntity> entityList = new ArrayList<>(this.entities);
        this.centerList = new ArrayList<>(this.clusterSize);
        this.clusterEntitiesList = new HashMap<>(this.clusterSize);

        for (int i = 0; i < clusterSize; i++)
            centerList.add(entityList.get(0));

        if (getAgentType() == StandardEntityURN.AMBULANCE_TEAM ||
                getAgentType() == StandardEntityURN.FIRE_BRIGADE ||
                getAgentType() == StandardEntityURN.POLICE_FORCE) {
            for (int i = 0; i < clusterSize; i++) {
                Human h = (Human)agents.get(i);
                StandardEntity centerEntity = worldInfo.getEntity(h.getPosition());
                if (centerList.contains(centerEntity)) {
                    do {
                        centerEntity = entityList.get(random.nextInt(entityList.size()));
                    } while (centerList.contains(centerEntity));
                }
                centerList.set(i, centerEntity);
            }
        } else {
            for (int i = 0; i < clusterSize; i++) {
                StandardEntity centerEntity = null;
                do {
                    centerEntity = entityList.get(random.nextInt(entityList.size()));
                } while (centerList.contains(centerEntity));
                centerList.set(i, centerEntity);
            }
        }

        for (int i = 0; i < repeat; i++) {
            this.clusterEntitiesList.clear();
            for (int index = 0; index < this.clusterSize; index++) {
                this.clusterEntitiesList.put(index, new ArrayList<>());
            }
            for (StandardEntity entity : entityList) {
                StandardEntity tmp = this.getNearEntityByLine(this.worldInfo, this.centerList, entity);
                this.clusterEntitiesList.get(this.centerList.indexOf(tmp)).add(entity);
            }
            for (int index = 0; index < this.clusterSize; index++) {
                int sumX = 0, sumY = 0;
                for (StandardEntity entity : this.clusterEntitiesList.get(index)) {
                    Pair<Integer, Integer> location = this.worldInfo.getLocation(entity);
                    sumX += location.first();
                    sumY += location.second();
                }
                int centerX = sumX / this.clusterEntitiesList.get(index).size();
                int centerY = sumY / this.clusterEntitiesList.get(index).size();
                StandardEntity center = this.getNearEntityByLine(this.worldInfo, this.clusterEntitiesList.get(index), centerX, centerY);
                if(center instanceof Area) {
                    this.centerList.set(index, center);
                }
                else if(center instanceof Human) {
                    this.centerList.set(index, this.worldInfo.getEntity(((Human) center).getPosition()));
                }
                else if(center instanceof Blockade) {
                    this.centerList.set(index, this.worldInfo.getEntity(((Blockade) center).getPosition()));
                }
            }
            if  (scenarioInfo.isDebugMode()) { System.out.print("*"); }
        }

        if  (scenarioInfo.isDebugMode()) { System.out.println(); }

        this.clusterEntitiesList.clear();
        for (int index = 0; index < this.clusterSize; index++) {
            this.clusterEntitiesList.put(index, new ArrayList<>());
        }
        for (StandardEntity entity : entityList) {
            StandardEntity tmp = this.getNearEntityByLine(this.worldInfo, this.centerList, entity);
            this.clusterEntitiesList.get(this.centerList.indexOf(tmp)).add(entity);
        }


        if(this.assignAgentsFlag) {
            assignAgents(worldInfo, new ArrayList<>(agents));
        }

        this.centerIDs = new ArrayList<>();
        for(int i = 0; i < this.centerList.size(); i++) {
            this.centerIDs.add(i, this.centerList.get(i).getID());
        }
        clusterEntityIDsList.clear();
        for (int index = 0; index < this.clusterSize; index++) {
            List<StandardEntity> entities = this.clusterEntitiesList.get(index);
            List<EntityID> list = new ArrayList<>(entities.size());
            for(int i = 0; i < entities.size(); i++) {
                list.add(i, entities.get(i).getID());
            }
            this.clusterEntityIDsList.add(list);
        }
    }

    protected void connectAllClusterEntities() {
        for (Map.Entry<Integer,List<StandardEntity>> entry : clusterEntitiesList.entrySet()) {
            List<StandardEntity> cEntities = entry.getValue();
            List<StandardEntity> newClusterEntities = connectEntities(cEntities);
            clusterEntitiesList.put(entry.getKey(), newClusterEntities);
        }
        updateClusterEntityIDList();
    }

    protected void updateClusterEntityIDList() {
        clusterEntityIDsList.clear();
        for (int i = 0; i < clusterEntitiesList.size(); i++) {
            List<StandardEntity> entities = clusterEntitiesList.get(i);
            clusterEntityIDsList.add(convertor.getEntityIDList(entities));
        }
    }

    public List<StandardEntity> getNeighbors(Area area, WorldInfo wi) {
        List<EntityID> nIDs = area.getNeighbours();
        if (nIDs != null) {
            List<StandardEntity> neighbors = new ArrayList<>(nIDs.size());
            for (EntityID nID : nIDs) {
                neighbors.add(wi.getEntity(nID));
            }
            return neighbors;
        }
        return new ArrayList<StandardEntity>(1);
    }

    protected List<StandardEntity> connectEntities(List<StandardEntity> cEntities) {
        List<List<StandardEntity>> paths = new ArrayList<>();
        Set<StandardEntity> entitySet = new HashSet<>(cEntities);
        Queue<StandardEntity> queue = new ArrayDeque<>(entitySet.size());
        Set<StandardEntity> visited = new HashSet<>(entitySet.size());
        while (!entitySet.isEmpty()) {
            List<StandardEntity> path = new ArrayList<>();
            queue.add(Iterables.get(entitySet, 0));
            while (!queue.isEmpty()) {
                StandardEntity current = queue.poll();
                visited.add(current);
                entitySet.remove(current);
                path.add(current);
                boolean hasNeighbor = false;
                if (current instanceof Area) {
                    Area currentArea = (Area)current;
                    List<StandardEntity> neighbors = this.getNeighbors(currentArea, worldInfo);
                    for (StandardEntity n : neighbors) {
                        if (!visited.contains(n) && entitySet.contains(n)) {
                            visited.add(n);
                            queue.add(n);
                            hasNeighbor = true;
                            entitySet.remove(n);
                        }
                    }
                }
            }
            if (path.size() > 0) {
                paths.add(path);
            }
        }


        while(paths.size() > 1) {
            List<StandardEntity> longest = paths.get(0);
            List<StandardEntity> secondLongest = paths.get(1);

            if (secondLongest.size() > longest.size()) {
                List<StandardEntity> temp = longest;
                longest = secondLongest;
                secondLongest = temp;
            }
            for (int i = 2; i < paths.size(); i++) {
                List<StandardEntity> path = paths.get(i);
                if (path.size() > longest.size()) {
                    longest = path;
                } else if (path.size() > secondLongest.size()) {
                    secondLongest = path;
                }
            }

            List<EntityID> shortestPath = null;
            int shortestPathLength = Integer.MAX_VALUE;
            for (int i = 0; i < secondLongest.size(); i++) {
                List<EntityID> path = pathPlanning.setFrom(secondLongest.get(i).getID()).setDestination( convertor.getEntityIDList(longest) ).calc().getResult();
                if (path != null) {
                    if (path.size() < shortestPathLength) {
                        shortestPathLength = path.size();
                        shortestPath = path;
                    }
                }
            }

            for (int i = 0; i < shortestPath.size(); i++) {
                longest.add(worldInfo.getEntity(shortestPath.get(i)));
            }
            for (int i = 0; i < secondLongest.size(); i++) {
                longest.add(secondLongest.get(i));
            }

            paths.remove(secondLongest);
        }

        if (paths.size() > 0) {
            return paths.get(0);
        }
        else {
            return new ArrayList<StandardEntity>();
        }
    }

    private void assignAgents(WorldInfo world, List<StandardEntity> agentList) {
        agentIDClusterIndexMapping.clear();
        int clusterIndex = 0;
        while (agentList.size() > 0) {
            StandardEntity center = this.centerList.get(clusterIndex);
            StandardEntity agent = this.getNearAgent(world, agentList, center);
            agentList.remove(agent);
            agentIDClusterIndexMapping.put(agent.getID(), clusterIndex);
            clusterIndex++;
            if (clusterIndex >= this.clusterSize) {
                clusterIndex = 0;
            }
        }
    }

    private StandardEntity getNearEntityByLine(WorldInfo world, List<StandardEntity> srcEntityList, StandardEntity targetEntity) {
        Pair<Integer, Integer> location = world.getLocation(targetEntity);
        return this.getNearEntityByLine(world, srcEntityList, location.first(), location.second());
    }

    private StandardEntity getNearEntityByLine(WorldInfo world, List<StandardEntity> srcEntityList, int targetX, int targetY) {
        StandardEntity result = null;
        for(StandardEntity entity : srcEntityList) {
            result = ((result != null) ? this.compareLineDistance(world, targetX, targetY, result, entity) : entity);
        }
        return result;
    }

    private StandardEntity getNearAgent(WorldInfo worldInfo, List<StandardEntity> srcAgentList, StandardEntity targetEntity) {
        List<EntityID> targetIDs = new ArrayList<>();
        for (StandardEntity entity : srcAgentList) {
            targetIDs.add(((Human)entity).getPosition());
        }
        pathPlanning.setFrom(targetEntity.getID());
        pathPlanning.setDestination(targetIDs);
        List<EntityID> shortestPath = pathPlanning.calc().getResult();
        EntityID agentPosition = shortestPath.get(shortestPath.size()-1);
        int index = targetIDs.indexOf(agentPosition);
        return srcAgentList.get(index);
    }




    private double getDistance(double fromX, double fromY, double toX, double toY) {
        double dx = fromX - toX;
        double dy = fromY - toY;
        return Math.hypot(dx, dy);
    }


    private StandardEntity compareLineDistance(WorldInfo worldInfo, int targetX, int targetY, StandardEntity first, StandardEntity second) {
        Pair<Integer, Integer> firstLocation = worldInfo.getLocation(first);
        Pair<Integer, Integer> secondLocation = worldInfo.getLocation(second);
        double firstDistance = getDistance(firstLocation.first(), firstLocation.second(), targetX, targetY);
        double secondDistance = getDistance(secondLocation.first(), secondLocation.second(), targetX, targetY);
        return (firstDistance < secondDistance ? first : second);
    }

    protected StandardEntity comparePathDistances(WorldInfo worldInfo, List<StandardEntity> targets, StandardEntity source) {
       List<EntityID> targetIDs = new ArrayList<>();
       for (StandardEntity entity : targets) {
           targetIDs.add(entity.getID());
       }
       pathPlanning.setFrom(source.getID());
       pathPlanning.setDestination(targetIDs);
       List<EntityID> shortestPath = pathPlanning.calc().getResult();
       EntityID shortestTargetID = shortestPath.get(shortestPath.size()-1);
       return worldInfo.getEntity(shortestTargetID);
    }

    protected int calPathLength(Map<EntityID, EntityID> ancestors, EntityID start, EntityID target) {
        int count = 0;
        EntityID current = target;
        do {
            count++;
            current = ancestors.get(current);
            if (current == null) throw new RuntimeException("Found a node with no ancestor! Something is broken.");
        } while (current != start);
        return count;
    }

    private void initShortestPath(WorldInfo worldInfo) {
        Map<EntityID, Set<EntityID>> neighbours = new LazyMap<EntityID, Set<EntityID>>() {
            @Override
            public Set<EntityID> createValue() {
                return new HashSet<>();
            }
        };
        for (Entity next : worldInfo) {
            if (next instanceof Area) {
                Collection<EntityID> areaNeighbours = ((Area) next).getNeighbours();
                neighbours.get(next.getID()).addAll(areaNeighbours);
            }
        }
        this.shortestPathGraph = neighbours;
    }

    private List<EntityID> shortestPath(EntityID start, Collection<EntityID> goals) {
        List<EntityID> open = new LinkedList<>();
        Map<EntityID, EntityID> ancestors = new HashMap<>();
        open.add(start);
        EntityID next;
        boolean found = false;
        ancestors.put(start, start);
        do {
            next = open.remove(0);
            if (isGoal(next, goals)) {
                found = true;
                break;
            }
            Collection<EntityID> neighbours = shortestPathGraph.get(next);
            if (neighbours.isEmpty()) continue;

            for (EntityID neighbour : neighbours) {
                if (isGoal(neighbour, goals)) {
                    ancestors.put(neighbour, next);
                    next = neighbour;
                    found = true;
                    break;
                }
                else if (!ancestors.containsKey(neighbour)) {
                    open.add(neighbour);
                    ancestors.put(neighbour, next);
                }
            }
        } while (!found && !open.isEmpty());
        if (!found) {
            return null;
        }
        EntityID current = next;
        List<EntityID> path = new LinkedList<>();
        do {
            path.add(0, current);
            current = ancestors.get(current);
            if (current == null) throw new RuntimeException("Found a node with no ancestor! Something is broken.");
        } while (current != start);
        return path;
    }

    private boolean isGoal(EntityID e, Collection<EntityID> test) {
        return test.contains(e);
    }

    protected StandardEntityURN getAgentType() {
        StandardEntity se = worldInfo.getEntity(agentInfo.getID());
        if (se instanceof PoliceForce) {
            return StandardEntityURN.POLICE_FORCE;
        } else if (se instanceof FireBrigade) {
            return StandardEntityURN.FIRE_BRIGADE;
        } else if (se instanceof AmbulanceTeam) {
            return StandardEntityURN.AMBULANCE_TEAM;
        } else if (se instanceof PoliceOffice) {
            return StandardEntityURN.POLICE_OFFICE;
        } else if (se instanceof FireStation) {
            return StandardEntityURN.FIRE_STATION;
        } else if (se instanceof AmbulanceCentre) {
            return StandardEntityURN.AMBULANCE_CENTRE;
        } else {
            return null;
        }
    }
}